#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
File Name: sequence2structure.py
Author: Dingfeng Wu
Creator: Dingfeng Wu
Date Created: 2024-01-03
Last Modified: 2023-01-05
Version: 1.0.0
License: MIT License
Description: Sequence-to-Structure (Seq2struc) is a computing process based on RDkit and the characteristics of cyclic peptide sequences, which can create cyclic peptide sequecne and convert sequence to cyclic peptide SMILES.

Copyright Information: Copyright (c) 2023 dfwlab (https://dfwlab.github.io/)

The code in this script can be used under the MIT License.
"""

import pandas as pd
from IPython.display import SVG, display
from rdkit import Chem
from rdkit.Chem.Draw import rdMolDraw2D
import re
from .setting import *

def transfrom(seequence, references, cyclic=True):#没写完
    Essential_AA = {'Ala', 'Arg', 'Asn', 'Asp', 'Cys', 'Glu', 'Gln', 'Gly', 'His', 'Ile', 'Leu', 'Lys',
    'Met', 'Phe', 'Pro', 'Trp', 'Tyr', 'Val', 'Ser', 'Thr'}

def seq2stru_essentialAA(sequence, cyclic=True):

    """
    ``"cyclic"`` controls whether it is looped, and when ``"cyclic = True"``, it generates a looped structure, otherwise a chain.
    
    Exmaple1::

        smiles, peptide = sequence2structure.seq2stru_essentialAA(sequence='APG', cyclic=False)
        print('SMILES:', smiles)
        basic.plot_molecule(peptide, w=300, h=300, isdisplay=True)

    SMILES: C[C@H](N)C(=O)N1CCC[C@H]1C(=O)NCC(=O)O
    
    .. image:: s2s.png
        :alt: Description of the image
        :align: center
        :scale: 50%


    :py:func:`basic.plot_molecule` function is used to output the resulting structure, or you can use :py:func:`basic.plot_smiles` .

    Example2::

        smiles, peptide = sequence2structure.seq2stru_essentialAA(sequence='APG', cyclic=True)
        print('SMILES:', smiles)
        basic.plot_smiles(smiles, w=300, h=300, isdisplay=True)
        
    SMILES: C[C@H](N)C(=O)N1CCC[C@H]1C(=O)NCC(=O)O

    .. image:: s2s.png
        :alt: Description of the image
        :align: center
        :scale: 50%

    Enter a sequence in 'Ala-Ala-Cys-Asp' format that will be converted to a single-letter structure by the :py:func:`code2symbol` function.

    Example3::

        smiles, peptide = sequence2structure.seq2stru_essentialAA(sequence='Ala-Ala-Cys-Asp', cyclic=True)
        print('SMILES:', smiles)
        
    SMILES: C[C@@H]1NC(=O)[C@H](C)NC(=O)[C@H](CC(=O)O)NC(=O)[C@H](CS)NC1=O

    """
    try:
        if '-' in sequence:
            sequence = [code2symbol(code)[0] for code in sequence.split('-')]
            sequence = ''.join(sequence)
        peptide = create_peptide_of_essentialAA(sequence, cyclic=cyclic)
        peptide = Chem.RemoveHs(peptide)  # 移除所有隐式氢原子
        Chem.AssignAtomChiralTagsFromStructure(peptide)  # 重新计算所有隐式氢原子
        smiles = Chem.MolToSmiles(peptide, canonical=True)
        return smiles, peptide
    except:
        return None, None


def seq2stru_no_essentialAA(sequence, references, cyclic=True):
    """
    Example::

        sequence = 'Aad--NMe-Ala--4OH-Thr--3Me-Pro'
        smiles, peptide = sequence2structure.seq2stru_no_essentialAA(sequence, references, cyclic=True)
        print(smiles)
        basic.plot_molecule(peptide, w=300, h=300, isdisplay=True)

    SMILES : CC1CCN2C(=O)C(C(O)CO)NC(=O)C(C)N(C)C(=O)C(CCCC(=O)O)NC(=O)C12

    .. image:: ns2s.png
        :alt: Description of the image
        :align: center
        :scale: 50%

    """
    try:
        sequence = [i.strip() for i in re.sub(r'\(\d+\)', '', sequence).split('--')]
        peptide = create_peptide_of_no_essentialAA(sequence, references, cyclic=cyclic)
        smiles = Chem.MolToSmiles(peptide, canonical=True)
        try:
            peptide = Chem.RemoveHs(peptide)  # 移除所有隐式氢原子
            Chem.AssignAtomChiralTagsFromStructure(peptide)  # 重新计算所有隐式氢原子
            return smiles, peptide
        except:
            return smiles, peptide
    except:
        return None, None


def code2symbol(code):
    """

    :param code: sequence.
    :type code: amino acid chain format
    :return: One letter code

    """

    # 打开文本文件并读取内容
    file_path = AminoAcids_path  # 替换为你的文件路径
    AminoAcids = []
    with open(file_path, 'r', encoding='utf-8') as file:
        """Reads the contents of the file line by line and adds it to the list"""
        for line in file:
            """# 去除行尾的换行符，并将参数添加到列表中"""
            AminoAcids.append(eval(line.strip()))
    c2s = {i[1]: [i[0], i[2]] for i in AminoAcids}
    if code.upper() in c2s.keys():
        return c2s[code.upper()]
    else:
        return None


def link_aa_by_peptide_bond(mol, c_index, n_index):
    o_index = None
    h_index = None
    # 查找与该碳原子相连的羟基的氧和氢原子
    for atom in mol.GetAtomWithIdx(c_index).GetNeighbors():
        if atom.GetSymbol() == 'O' and mol.GetBondBetweenAtoms(atom.GetIdx(), c_index).GetBondType() == Chem.BondType.SINGLE:
            o_index = atom.GetIdx()
            for h_atom in atom.GetNeighbors():
                if h_atom.GetAtomicNum() == 1:  # 氢原子
                    h_index = h_atom.GetIdx()
    # print(o_index, h_index)
    # 创建一个可编辑的分子
    emol = Chem.EditableMol(mol)
    # 然后再删除羟基氧原子和氢原子
    emol.RemoveAtom(h_index) if h_index else 1
    emol.RemoveAtom(o_index) if o_index else 1
    emol.AddBond(c_index, n_index, order=Chem.rdchem.BondType.SINGLE)
    # 获取修改后的分子
    mol = emol.GetMol()
    return mol

def detect_backbone(mol):
    # 识别骨架并将骨架分子编号为0至N，其他侧链分子编号大于N
    # renumber the backbone atoms so the sequence order is correct:
    # N端乙酰化和C端酰胺化: https://zhuanlan.zhihu.com/p/540769025
    # 肽链一般从N端读到C端,C端一般保留一个COOH,因此C端为HO-C(=O),中间一个C,然后加一个肽键NC(=O),后续就都是C和肽键，一直到N端
    # 正反识别都可以，反向就是OC(=O)CNC(=O)CNC(=O)CN...,如果C端酰胺化,则为NC(=O)CNC(=O)CNC(=O)CN...
    # N端不管有没有乙酰化,都会保留N,无乙酰化为-NH2,有乙酰化为-NC(=O)CH3;
    bbsmiles = "C(=O)CN" * len(
        mol.GetSubstructMatches(Chem.MolFromSmiles('NCC=O')))  # 使用GLY最小氨基酸单元识别氨基酸数量，生成骨架generate backbone SMILES
    # backbone = mol.GetSubstructMatches(Chem.MolFromSmiles(bbsmiles))[0] 报错tuple index out of range
    matches = mol.GetSubstructMatches(Chem.MolFromSmiles(bbsmiles))
    backbone = ''
    if matches:  # 检查是否有匹配结果
        backbone = matches[0]
        # 其他操作
    else:
        print("No matches found.")
        return backbone, None # 如果找不到肽键，或者数量不同则退出，比如存在两个线性肽链通过两个链接形成环的情况，需要逐一搜索；
    backbone_idx = list(backbone)
    backbone_idx.reverse()
    return backbone, backbone_idx
 

def create_peptide_of_essentialAA(sequence, cyclic=True):
    """

    :param sequence: sequence of essential AA.
    :type sequence: One letter code
    :param cyclic: Optional "cyclic".
    :type cyclic: True or False
    :return: mol

    """

    try:
        mol = Chem.MolFromSequence(sequence)
    except:
        return None
    if cyclic:
        backbone, backbone_idx = detect_backbone(mol)
        c_index = backbone[0]
        n_index = backbone[-1]
        mol = link_aa_by_peptide_bond(mol, c_index, n_index)
    return mol


def reference_aa_monomer(monomers_path=monomer_path): #未修改
    """

    :param monomers_path: sequence of essential AA.
    :type sequence: One letter code
    :param cyclic: Optional "cyclic".
    :type cyclic: True or False
    :return: mol

    """
    monomers = pd.read_csv(monomers_path, sep='\t', index_col=0)
    references = {}
    for i in monomers.index:
        smile = monomers.loc[i, 'Smiles']
        code = monomers.loc[i, 'Code']
        weight = float(monomers.loc[i, 'Weight'])
        symbol = monomers.loc[i, 'Symbol']
        symbol = code if str(symbol).strip() == '' or str(symbol).upper().strip() == 'NAN' else symbol
        try:
            aa = Chem.MolFromSmiles(smile)
            match = aa.GetSubstructMatches(Chem.MolFromSmiles("NCC=O"))
            if match and match[0]:
                num_atoms = aa.GetNumAtoms()
                references[code] = [aa, symbol, num_atoms, match[0]]
        except:
            pass
    return references

def create_peptide_of_no_essentialAA(sequence, references, cyclic=True):
    """

    :param sequence: sequence of noessential AA.
    :type sequence: One letter code
    :param references: monomers.
    :type references: from :py:func:`reference_aa_monomer`
    :param cyclic: Optional "cyclic".
    :type cyclic: True or False
    :return: mol

    """
    peptide, pep_idx = references[sequence[0]][0], references[sequence[0]][-1]
    pep_c_index, first_n_index = pep_idx[2], pep_idx[0]
    for aa_code in sequence[1:]:
        aa, idx = references[aa_code][0], references[aa_code][-1]
        c_index, n_index = idx[2], idx[0]
        peptide, pep_c_index = connect_two_aa_with_peptide_bond(peptide, pep_c_index, aa, c_index, n_index)
    if cyclic:
        peptide = link_aa_by_peptide_bond(peptide, pep_c_index, first_n_index)
    return peptide

def connect_two_aa_with_peptide_bond(aa1, c_index1, aa2, c_index2, n_index2):
    c_index2, n_index2 = c_index2 + aa1.GetNumAtoms() - 1, n_index2 + aa1.GetNumAtoms() - 1
    peptide = Chem.CombineMols(aa1, aa2)
    peptide = link_aa_by_peptide_bond(peptide, c_index1, n_index2)
    return peptide, c_index2


def get_molblock(mol):
    return Chem.MolToMolBlock(mol, forceV3000=True)


def get_molblock_from_smiles(smiles):
    mol = Chem.MolFromSmiles(smiles)
    return Chem.MolToMolBlock(mol, forceV3000=True)



import re
from .setting import *

# 打开文本文件并读取内容
file_path = AminoAcids_path  # 替换为你的文件路径
AminoAcids = []
with open(file_path, 'r', encoding='utf-8') as file:
    # 逐行读取文件内容并添加到列表中
    for line in file:
        # 去除行尾的换行符，并将参数添加到列表中
        AminoAcids.append(eval(line.strip()))

################## 序列格式互转 ##################
def read_sequence(sequence):

    """
    **read_sequence** reads sequences in a variety of formats, converting the item into nodes(lists of amino acids) and edges(link information).
    you can also use exact function to transform, such as :py:func:`read_iupac_condensed`, :py:func:`read_graph_representation`, and :py:func:`read_one_letter_sequence`

    :param sequence: cyclic peptide sequence.
    :type sequence: *Format* [Graph presentation, IUPAC condensed, Amino acid chain, One letter code]
    :return: seq_format, nodes, edges
    :rtype: string, List[str], List[(int, int)]

    Example::

        seq_format, nodes, edges = SequenceTransformer.read_sequence('aThr,Tyr,dhAbu,bOH-Gln,Gly,Gln,His,Dab,C13:2(t4.t6)-OH(2.3),Lyx,dhAbu @1,5 @6,10 @0,8')
        print(seq_format, nodes, edges)

    Results::
    
        Graph presentation\n
        ['aThr', 'Tyr', 'dhAbu', 'bOH-Gln', 'Gly', 'Gln', 'His', 'Dab', 'C13:2(t4.t6)-OH(2.3)', 'Lyx', 'dhAbu']\n
        [(1, 5), (6, 10), (0, 8)]

    """

    seq_format = ''
    if '@' in sequence:
        seq_format = 'Graph presentation'
        nodes, edges = read_graph_representation(sequence)
    elif ',' in sequence:
        no_parentheses = re.sub(r'\([^)]*\)', '', sequence)  # 删除所有括号中的逗号
        if ',' in no_parentheses:
            seq_format = 'Graph presentation'
            nodes, edges = read_graph_representation(sequence)
        else:
            seq_format = 'IUPAC condensed'
            nodes, edges = read_iupac_condensed(sequence, sep='-')
    elif '--' in sequence:
        seq_format = 'Amino acid chain'
        nodes, edges = read_iupac_condensed(sequence, sep='--')
    elif '-' in sequence:
        try:
            seq_format = 'IUPAC condensed'
            nodes, edges = read_iupac_condensed(sequence, sep='-')
        except:
            nodes, edges = '', ''
    else:
        seq_format = 'One letter peptide'
        nodes, edges = read_one_letter_sequence(sequence)
    return seq_format, nodes, edges


def create_sequence(nodes, edges):
    """
    **create_sequence** can generate processed nodes(amino acid lists) and edges(link information) into sequences in many different formats.
    you can also use exact function to create, such as :py:func:`create_iupac_condensed`, :py:func:`create_graph_presentation`, :py:func:`create_amino_acid_chain`, and :py:func:`create_one_letter_peptide`

    :param nodes: lists of amino acids.
    :param edges: link information.
    :type edges: List[(int, int)]
    :return: {'iupac_condensed': iupac_condensed, 'amino_acid_chain': amino_acid_chain,
            'graph_presentation': graph_presentation, 'one_letter_peptide': one_letter_peptide}

    Example::

        seq_list = SequenceTransformer.create_sequence(['aThr', 'Tyr', 'dhAbu', 'bOH-Gln', 'Gly', 'Gln', 'His', 'Dab', 'C13:2(t4.t6)-OH(2.3)', 'Lyx', 'dhAbu'], [(1, 5), (6, 10), (0, 8)])
        print(seq_list)
    
    Results::

        {\n
        'iupac_condensed': 'aThr(3)-Tyr(1)-dhAbu-bOH-Gln-Gly-Gln(1)-His(2)-Dab-C13:2(t4.t6)-OH(2.3)(3)-Lyx-dhAbu(2)',\n
        'amino_acid_chain': 'aThr(3)--Tyr(1)--dhAbu--bOH-Gln--Gly--Gln(1)--His(2)--Dab--C13:2(t4.t6)-OH(2.3)(3)--Lyx--dhAbu(2)',\n
        'graph_presentation': 'aThr,Tyr,dhAbu,bOH-Gln,Gly,Gln,His,Dab,C13:2(t4.t6)-OH(2.3),Lyx,dhAbu @1,5 @6,10 @0,8',\n
        'one_letter_peptide': None\n
        }

    """
    try:
        iupac_condensed = create_iupac_condensed(nodes, edges)
    except:
        iupac_condensed = None
    try:
        amino_acid_chain = create_amino_acid_chain(nodes, edges)
    except:
        amino_acid_chain = None
    try:
        graph_presentation = create_graph_presentation(nodes, edges)
    except:
        graph_presentation = None
    try:
        one_letter_peptide = create_one_letter_peptide(nodes)
    except:
        one_letter_peptide = None
    return {'iupac_condensed': iupac_condensed, 'amino_acid_chain': amino_acid_chain,
            'graph_presentation': graph_presentation, 'one_letter_peptide': one_letter_peptide}


def read_iupac_condensed(sequence, sep='-'):
    """

    :param sequence: cyclic peptide sequence.
    :type sequence: IUPAC condensed
    :param sep: Optional "sep".
    :type sep: Amino acid link in sequences
    :return: amino acids, edges
    :rtype: List[str], List[(int, int)]


    """

    # 0. 确认是否有盐或多条链
    if '.' in sequence:
        sequence = sequence.split('.')[0].strip()
    # 1. 确认是否有cyclo[]成环信息
    iscyclo = True if 'cyclo' in sequence else False
    sequence = sequence.replace('cyclo', '').strip('[]')  # 移除 'cyclo' 和方括号
    sequence = re.sub(r'\([^)]*\)', replace_hyphen, sequence)  # 替换（）中的-为～～，（）中为修饰
    items = sequence.split(sep)
    # 2. 确认首尾修饰信息
    header = ''
    tail = ''
    if items[0] in ['H', 'NH2', 'Unk']:
        header = items[0]
        items = items[1:]
    if items[-1] in ['H', 'NH2', 'Unk']:
        tail = items[-1]
        items = items[:-1]
    # 3. 获取氨基酸，DL-修饰添加进氨基酸，首位修饰添加进氨基酸
    amino_acids = []
    edge_marks = []
    pred = ''
    max_edge_mark = 0
    for item in items:
        if item in ['DL', 'D', 'L']:
            pred = item + '-'
            continue
        amino_acid = (pred + item).replace('~~', '-').strip()
        edge_mark = []
        while True:
            edge_mark_match = re.search(r'\(\d+\)$', amino_acid)
            if edge_mark_match:
                amino_acid = amino_acid.replace(edge_mark_match[0], '').strip()
                em = int(edge_mark_match[0][1:-1])
                max_edge_mark = max([max_edge_mark, em])
                edge_mark.append(em)
            else:
                break
        amino_acids.append(amino_acid)
        edge_marks.append(edge_mark)
        pred = ''
    if header:
        amino_acids[0] = header + '-' + amino_acids[0]
    if tail:
        amino_acids[-1] = amino_acids[-1] + '-' + tail
    # 4. 处理最后空格但有edge标签的
    if amino_acids[-1] == '' and edge_marks:
        amino_acids = amino_acids[:-1]
        edge_mark = edge_marks[-1][:]
        edge_marks = edge_marks[:-1]
        edge_marks[-1].extend(edge_mark)
    # 5. 处理氨基酸N(1)标注
    for i, aa in enumerate(amino_acids):
        match = re.search(r'N\((\d)\)', aa)
        if match:
            amino_acids[i] = re.sub(r'N\((\d)\)', 'N-', aa)
            edge_marks[i].append(int(match.group(1)))

    # print(amino_acids)
    # print(edge_marks)
    # 6. edges信息处理
    edges = []
    for i in range(1, max_edge_mark + 1):
        pair = []
        for idx in range(len(edge_marks)):
            if i in edge_marks[idx]:
                pair.append(idx)
        edges.append((pair[0], pair[1]))
    if iscyclo:
        edges.append((0, len(amino_acids) - 1))
    return amino_acids, edges


def read_graph_representation(sequence):
    """

    :param sequence: cyclic peptide sequence.
    :type sequence: graph representation
    :return: amino acids, edges
    :rtype: List[str], List[(int, int)]

    """

    amin_acids = [i.strip() for i in sequence.split(' @')[0].strip().split(',')]
    edges = [tuple([int(j) for j in i.strip().split(',')]) for i in sequence.split('@')[1:]]
    return amin_acids, edges


def read_one_letter_sequence(sequence):
    """

    :param sequence: cyclic peptide sequence.
    :type sequence: one letter code
    :return: amino acids, []
    :rtype: List[str], []

    """
    tail = ''
    if '(NH2)' == sequence.upper()[-5:]:
        sequence = sequence[:-5]
        tail = '(NH2)'
    amino_acid_refs = {i: j for i, j, _, _ in AminoAcids}
    amino_acid_refs['X'] = 'UNK'
    aas = [amino_acid_refs[i] for i in sequence]
    aas[-1] = aas[-1] + tail
    return aas, []


def create_iupac_condensed(nodes, edges):
    """    
    
    :param nodes: lists of amino acids.
    :param edges: link information.
    :type edges: List[(int, int)]
    :return: sequence
    :rtype: IUPAC condensed
    
    """
    edge_marker = 1
    cyclo = False
    amino_acids = nodes[:]
    for i, j in edges:
        if min(i, j) == 0 and max(i, j) == len(amino_acids) - 1:
            cyclo = True
            continue
        amino_acids[i] += f'({edge_marker})'
        amino_acids[j] += f'({edge_marker})'
        edge_marker += 1
    sequence = '-'.join(amino_acids)
    sequence = f'cyclo[{sequence}]' if cyclo else sequence
    return sequence


def create_amino_acid_chain(nodes, edges):
    """    
    
    :param nodes: lists of amino acids.
    :param edges: link information.
    :type edges: List[(int, int)]
    :return: sequence
    :rtype: Amino acid chain
    
    """

    edge_marker = 1
    amino_acids = nodes[:]
    for i, j in edges:
        amino_acids[i] += f'({edge_marker})'
        amino_acids[j] += f'({edge_marker})'
        edge_marker += 1
    sequence = '--'.join(amino_acids)
    return sequence


def create_graph_presentation(nodes, edges):
    """    
    
    :param nodes: lists of amino acids.
    :param edges: link information.
    :type edges: List[(int, int)]
    :return: sequence
    :rtype: Graph presentation
    
    """
    sequence = ','.join(nodes)
    edges_pre = '' if len(edges) == 0 else ' '.join(['@' + ','.join([str(k) for k in pair]) for pair in edges])
    return (sequence + ' ' + edges_pre).strip()


def create_one_letter_peptide(nodes):  # 忽略侧链互作
    """    
    
    :param nodes: lists of amino acids.
    :param edges: link information.
    :type edges: List[(int, int)]
    :return: sequence
    :rtype: one letter peptide
    
    """

    amino_acid_refs = {i: j for j, i, _, _ in AminoAcids}
    amino_acid_refs['UNK'] = 'X'
    tail = ''
    if '(NH2)' == nodes[-1].upper()[-5:]:
        nodes[-1] = nodes[-1][:-5]
        tail = '(NH2)'
    is_essential_aa = [i.upper().strip() in amino_acid_refs.keys() for i in nodes]
    if False in is_essential_aa:
        return None
    else:
        return ''.join([amino_acid_refs[i.upper().strip()] for i in nodes]) + tail


def replace_hyphen(match):
    # 替换括号内的所有 hyphen (-) 为 tilde (~)
    return match.group(0).replace('-', '~~')


from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import AllChem, Descriptors, rdMolDescriptors, Lipinski, rdFingerprintGenerator, \
    MACCSkeys, RDKFingerprint
from rdkit.Chem.rdMolDescriptors import CalcTPSA

logger = RDLogger.logger()
logger.setLevel(RDLogger.CRITICAL)  # 只显示 CRITICAL 级别的日志

################## 化学性质 ##################
def chemial_physical_properties_from_smiles(smiles):
    """
    Calculate the cyclopeptide chemical property from SMILES.

    :param smiles: Cyclic peptides SMILES information.
    :return: chemial and physical properties
    :rtype: dict{property name: value}

    you can also use :py:func:`cal_chemial_physical_properties` to calculate the properties from mol

    List of the chemical properties:

    .. csv-table:: 
        :file: Chemical_P.csv
        :header: "Property", "Description"

    """

    try:
        mol = Chem.MolFromSmiles(smiles)
    except:
        return {}
    properties = cal_chemial_physical_properties(mol)
    properties = cal_rules(properties)
    properties['RDKit_Fingerprint'] = cal_RDKit_fingerprint(mol)
    properties['Daylight_like_Fingerprint'] = cal_daylight_like_fingerprint(mol)
    properties['Morgan_Fingerprint'] = cal_morgan_fingerprint(mol)
    properties['MACCS_Keys'] = cal_MACCS_keys(mol)
    return properties


def cal_chemial_physical_properties(mol):
    # 理化性质函数
    properties = {
        'Exact_Mass': Descriptors.ExactMolWt(mol),
        'Topological_Polar_Surface_Area': CalcTPSA(mol),
        'Complexity': Descriptors.FpDensityMorgan1(mol),
        'Crippen_LogP': Descriptors.MolLogP(mol),  # 计算LogP Crippen
        'Heavy_Atom_Count': Lipinski.HeavyAtomCount(mol),
        'Hydrogen_Bond_Donor_Count': Lipinski.NumHDonors(mol),
        'Hydrogen_Bond_Acceptor_Count': Lipinski.NumHAcceptors(mol),
        'Rotatable_Bond_Count': Lipinski.NumRotatableBonds(mol),
        'Formal_Charge': Chem.GetFormalCharge(mol),
        'Refractivity': Descriptors.MolMR(mol),
        'Number_of_Rings': rdMolDescriptors.CalcNumRings(mol),
        'Number_of_Atoms': mol.GetNumAtoms(),
    }
    return properties


def cal_rules(properties):
    """
    
    :param properties: A list of calculated properties.
    :type properties: calculated by :py:func:`chemial_physical_properties_from_smiles`
    :return: properties

    .. list-table::
        :header-rows: 1

        * - Property
          - Description
        * - Rule of Five
          - Indicates non-compliance with Lipinski's rule of five, \n
            suggesting potential issues with bioavailability as an oral drug.
        * - Veber's Rule
          - Shows non-adherence to Veber's rules, \n
            potentially impacting oral bioavailability and permeability.
        * - Ghose Filter
          - A molecular property filter used to assess the \n
            drug-likeness of a compound based on its physicochemical properties.

     """
    # Lipinski 规则五
    lipinski_rule_of_five = properties['Hydrogen_Bond_Donor_Count'] <= 5 and \
                            properties['Hydrogen_Bond_Acceptor_Count'] <= 10 and \
                            properties['Exact_Mass'] <= 500 and \
                            properties['Crippen_LogP'] <= 5

    # Veber 规则
    vebers_rule = properties['Topological_Polar_Surface_Area'] <= 140 and \
                  properties['Rotatable_Bond_Count'] <= 10

    # 检查 Ghose Filter 条件
    ghose_filter = 160 <= properties['Exact_Mass'] <= 480 and \
                   0.4 <= properties['Crippen_LogP'] <= 5.6 and \
                   20 <= properties['Number_of_Atoms'] <= 70

    properties['Rule_of_Five'] = lipinski_rule_of_five
    properties["Veber_Rule"] = vebers_rule
    properties['Ghose_Filter'] = ghose_filter
    return properties


################## 分子指纹 ##################

def cal_RDKit_fingerprint(mol):
    # 生成RDKit指纹
    fp = RDKFingerprint(mol)
    bit_vector = fp.ToBitString()
    return bit_vector

def cal_daylight_like_fingerprint(mol):
    # 生成拓扑指纹
    fp = rdFingerprintGenerator.GetFPs([mol])
    bit_vector = fp[0].ToBitString()
    return bit_vector


def cal_morgan_fingerprint(mol):
    """ Generate a Morgan fingerprint (radius 2) """
    # 生成Morgan指纹（半径为2）
    morgan_fp = AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=1024)
    bit_vector = morgan_fp.ToBitString()
    return bit_vector


def cal_MACCS_keys(mol):
    # 生成 MACCS keys 指纹
    maccs_fp = MACCSkeys.GenMACCSKeys(mol)
    bit_vector = maccs_fp.ToBitString()
    return bit_vector


def calculate_amino_acid_composition(sequence):
    # 打开文本文件并读取内容
    file_path = 'AminoAcids.txt'  # 替换为你的文件路径
    AminoAcids = []
    with open(file_path, 'r', encoding='utf-8') as file:
        """Reads the contents of the file line by line and adds it to the list"""
        for line in file:
            """# 去除行尾的换行符，并将参数添加到列表中"""
            AminoAcids.append(eval(line.strip()))

    aa_list = [i[0] for i in AminoAcids]
    counts = [(aa, sequence.count(aa) if aa in sequence else 0) for aa in aa_list]
    return counts




import os
global monomer_path, aas_path, AminoAcids_path, aa_smiles_path

path_of_this_file = os.path.abspath(__file__).replace('/setting.py', '')
monomer_path = os.path.join(path_of_this_file,'monomer.tsv')
aas_path = os.path.join(path_of_this_file,'aas.txt')
aa_smiles_path = os.path.join(path_of_this_file,'aa_smiles.txt')
AminoAcids_path = os.path.join(path_of_this_file,'AminoAcids.txt')


from rdkit import Chem
from rdkit.Chem import AllChem

################## 结构格式互转 ##################
# 自动识别并读人SMILES InChI Molblock SDFblock PDBblock
# https://chemistry.stackexchange.com/questions/34563/pubchem-inchi-smiles-and-uniqueness
# 以InChi=1S/…开头的InChi标识符是标准InChI。在标准InChI中，InChI标识符“对于任何移动氢原子的排列都必须相同”
# 以InChI=1/…开始为非标准InChI，包括一个以/f开头的额外层（fixed-hydrogen layer）。
# 标准InChI生成结构与SMILES不同，而非标准InChI生成结构与SMILES相同，非标准InChI生成生成代码：Chem.MolToInchi(mol1, options='/FixedH')
# 自动识别并读人SMILES InChI Molblock SDFblock PDBblock


# 生成 SMILES InChI InChIKey Molblock SDFblock PDBblock
def output_molecule(mol, pdbblock=None, conformation=None):
    """

    :param mol: mol file.
    :param pdbblock: Optional "conformation".
    :type pdbblock: pdbblock or None
    :param conformation: Optional "conformation".
    :type conformation: conformation or None
    :return: mol

    """

    smiles = Chem.MolToSmiles(mol, canonical=True, isomericSmiles=True)
    inchi = Chem.MolToInchi(mol, options='/FixedH')
    inchikey = Chem.InchiToInchiKey(inchi)
    molblock = Chem.MolToMolBlock(mol, includeStereo=True)
    # PDBblock需要通过空间结构保存手性, 因此其他格式转PDB时需要先创建3D构象;
    if pdbblock is None and conformation is not None:
        pdbblock = Chem.MolToPDBBlock(conformation)
    return {'smiles': smiles, 'inchi': inchi, 'inchikey': inchikey, 'molblock': molblock, 'pdbblock': pdbblock}


def predict_3d_conformation(mol):
    """
    Predict the 3D structure
    """
    # molecule = Chem.MolFromSmiles(smiles)
    mol_3d = Chem.AddHs(mol.__copy__())  # 添加氢
    AllChem.EmbedMolecule(mol_3d, AllChem.ETKDG())  # 生成 3D 坐标
    return mol_3d


def mol_optimize(mol):
    """
    Universall Force Field is used for energy minimization.
    """
    # 能量最小化：UFF (Universal Force Field)适用于小分子的力场；多肽建模和优化，专门的生物分子建模工具（如 AMBER、CHARMM 或 GROMACS）会更合适
    AllChem.UFFOptimizeMolecule(mol)  # 使用 UFF 力场进行优化
    return mol


def mol2molblock(mol):
    return Chem.MolToMolBlock(mol)


def mol2pdbblock(mol):
    return Chem.MolToPDBBlock(mol)